# -*- coding: utf-8 -*-

'''
Metrics file(:mod:`krakatoa.models._metrics`)
============================================================
'''

#============================================================
# Imports
#============================================================

from sklearn import metrics

#============================================================
# Metrics
#============================================================
#TODO: Criar modulo para metrics

def getScores(kind):

    regression_scores = {
        'explained_variance' : {'f' : metrics.explained_variance_score, 'name' : 'explained_variance'},
        'max_error' : {'f' : metrics.max_error, 'name' : 'max_error'},
        'mae' : {'f' : metrics.mean_absolute_error, 'name' : 'neg_mean_absolute_error'},
        'mse' : {'f' : metrics.mean_squared_error, 'name' : 'neg_mean_squared_error'},
        'rmse' : {'f' : metrics.mean_squared_error, 'name' : 'neg_root_mean_squared_error'}, #TODO: criar funcao RMSE,
        'msle' : {'f' : metrics.mean_squared_log_error, 'name' : 'neg_mean_squared_log_error'},
        'neg_median_absolute_error' : {'f' : metrics.median_absolute_error, 'name' : 'neg_median_absolute_error'},
        'r2' : {'f' : metrics.r2_score, 'name' : 'r2'},
        'poison' : {'f' : metrics.mean_poisson_deviance, 'name' : 'neg_mean_poisson_deviance'},
        'gama' : {'f' : metrics.mean_gamma_deviance, 'name' : 'neg_mean_gamma_deviance'},
        'neg_mean_absolute_percentage_error' : {'f' : metrics.mean_absolute_percentage_error, 'name' : 'neg_mean_absolute_percentage_error'},
        'd2_absolute_error_score' : {'f' : metrics.d2_absolute_error_score, 'name' : 'd2_absolute_error_score'},
        'd2_pinball_score' : {'f' : metrics.d2_pinball_score, 'name' : 'd2_pinball_score'},
        'd2_tweedie_score' : {'f' : metrics.d2_tweedie_score, 'name' : 'd2_tweedie_score'}
        }


    classification_scores = {
        'accuracy' : {'f' : metrics.accuracy_score, 'name' : 'accuracy'},
        'balanced_accuracy' : {'f' : metrics.balanced_accuracy_score, 'name' : 'balanced_accuracy'},
        'roc_auc' : {'f' : metrics.roc_auc_score, 'name' : 'roc_auc'}
        }

    if kind == "classification":
        scores = classification_scores
    else:
        scores = regression_scores
    
    return scores



# -*- coding: utf-8 -*-

'''
Data Evaluation (:mod:`krakatoa.future.evaluate`)
============================================================
'''

# ============================================================
# Imports
# ============================================================

import pandas as pd
import numpy as np

# ============================================================


def categoryPerFeature(df, columns, sort=True):
    n_categories_per_feature = df[columns].apply(lambda x: len(set(x)))

    if sort:
        n_categories_per_feature = n_categories_per_feature.sort_values(
            ascending=False)

    return n_categories_per_feature


def countNull(df):

    # result = ''

    # if mode == 'perc':
    #     result = (((df.isnull().sum() | df.eq('').sum())/df.shape[0])*100)
    # elif mode == 'count':
    #     result = (df.isnull().sum() | df.eq('').sum()).to_dict()

    return pd.DataFrame({
        'variable': list(df.columns),
        'perc': (((df.isnull().sum() | df.eq('').sum())/df.shape[0])*100).values,
        'count': (df.isnull().sum() | df.eq('').sum()).values
    })



# -*- coding: utf-8 -*-

'''
Quick Models(:mod:`krakatoa.models.quick`)
============================================================
'''

# ============================================================
# Imports
# ============================================================

from sklearn.model_selection import cross_validate


from ._getmodels import getModels
from ._metrics import getScores
from ._model_selection import getModel
import pandas as pd

import time
# ============================================================
# Regression Functions
# ============================================================


class Regressor():
    def __init__(self):

        self._regression_scores = getScores('regression')

    def _runModels(self, models, x, y, scoring, cv, model_selection='cross_validate', metrics=['r2'], model_selection_params={}):
        results = {'estimator': [], 'fit_time': [], 'score_time': []}

        # Model selection part
        res_model_selection = getModel(model=model_selection)

        # model_selection_function
        modelSelFunc = res_model_selection['f']
        method = res_model_selection['method']
        
        if method == 0:
            # For methods that returns x_train, x_test, y_train, y_test

            X_train, X_test, y_train, y_test = modelSelFunc(
                x, y, **model_selection_params)
            dict_metrics = getScores('regression')
            for model in models:
                results['estimator'].append(model['name'])
                # Fit time
                fit_start_time = time.perf_counter()
                estimator = model['estimator']
                estimator.fit(X_train, y_train)
                fit_end_time = time.perf_counter()

                results['fit_time'].append(fit_end_time - fit_start_time)

                y_predicted = estimator.predict(X_test)

                metric_start_time = time.perf_counter()
                for metric in metrics:
                    dict_metric = dict_metrics.get(metric, None)
                    if dict_metric == None:
                        break

                    res_metric = dict_metric['f'](y_test, y_predicted)
                    metric_name = f'{metric}_score'
                    if metric_name not in results.keys():
                        results[metric_name] = []
                    results[metric_name].append(res_metric)

                metric_end_time = time.perf_counter()
                metric_time = metric_end_time - metric_start_time
                results['score_time'] = metric_time

        elif method == 1:
            # Must improve implementation in order to use other KFolds methods such as Group and Stratified
            folds = modelSelFunc(**model_selection_params)
            for model in models:

                res = cross_validate(
                    model['estimator'], x, y, scoring=scoring, cv=folds, n_jobs=-1)
                results['estimator'].append(model['name'])

                for k, v in res.items():
                    if k not in results.keys():
                        results[k] = []

                    results[k].append(v.mean())

        elif method == 2:

            for model in models:

                res = modelSelFunc(
                    model['estimator'], x, y, scoring=scoring, cv=cv, n_jobs=-1)
                results['estimator'].append(model['name'])

                for k, v in res.items():
                    if k not in results.keys():
                        results[k] = []

                    results[k].append(v.mean())

        return results

    def _runModelsYield(self, models, x, y, scoring, cv, model_selection='cross_validate', metrics=['r2'], model_selection_params={}):

        results = {'estimator': [], 'fit_time': [], 'score_time': []}

        # Model selection part
        res_model_selection = getModel(model=model_selection)

        # model_selection_function
        modelSelFunc = res_model_selection['f']
        method = res_model_selection['method']
        if method == 0:
            # For methods that returns x_train, x_test, y_train, y_test

            X_train, X_test, y_train, y_test = modelSelFunc(
                x, y, **model_selection_params)
            dict_metrics = getScores('regression')
            for model in models:
                yield {
                    'status_code': 1,
                    'status': 'running',
                    'message': 'Training running',
                    'data': model['name']
                }

                results['estimator'].append(model['name'])
                # Fit time
                fit_start_time = time.perf_counter()
                estimator = model['estimator']
                estimator.fit(X_train, y_train)
                fit_end_time = time.perf_counter()

                results['fit_time'].append(fit_end_time - fit_start_time)

                y_predicted = estimator.predict(X_test)

                metric_start_time = time.perf_counter()
                for metric in metrics:
                    dict_metric = dict_metrics.get(metric, None)
                    if dict_metric == None:
                        break

                    res_metric = dict_metric['f'](y_test, y_predicted)
                    metric_name = f'{metric}_score'
                    if metric_name not in results.keys():
                        results[metric_name] = []
                    results[metric_name].append(res_metric)

                metric_end_time = time.perf_counter()
                metric_time = metric_end_time - metric_start_time
                results['score_time'] = metric_time

        elif method == 1:
            # Must improve implementation in order to use other KFolds methods such as Group and Stratified
            folds = modelSelFunc(**model_selection_params)
            for model in models:
                yield {
                    'status_code': 1,
                    'status': 'running',
                    'message': 'Training running',
                    'data': model['name']
                }

                res = cross_validate(
                    model['estimator'], x, y, scoring=scoring, cv=folds, n_jobs=-1)
                results['estimator'].append(model['name'])

                for k, v in res.items():
                    if k not in results.keys():
                        results[k] = []

                    results[k].append(v.mean())

        elif method == 2:

            for model in models:
                yield {
                    'status_code': 1,
                    'status': 'running',
                    'message': 'Training running',
                    'data': model['name']
                }

                res = modelSelFunc(
                    model['estimator'], x, y, scoring=scoring, cv=cv, n_jobs=-1)
                results['estimator'].append(model['name'])

                for k, v in res.items():
                    if k not in results.keys():
                        results[k] = []

                    results[k].append(v.mean())

        yield {
            'status_code': 0,
            'status': 'finished',
            'message': 'Training finished',
            'data': results
        }

    def buildModel(self, x, y, models=['randomForest'], selMode='type'):
    
        models = getModels(mode='regression',
                           modelClasses=models, selMode=selMode)

        model = models[0]
        
        _model = model['estimator'].fit(x, y)

        return _model
   
    def linearRegression(self, x, y, score=['r2'], cv=5, **kwargs):
        '''
        Quick linear regression models evaluation.

        Parameters
        ----------
        x : DATAFRAME
            Dataset Features.
        y : NUMPY ARRAY
            Target.
        score : LIST, optional
            Scoring metrics. The default is ['r2'].
        cv : INT, optional
            Number of kfolds. The default is 5.

        Returns
        -------
        return : DATAFRAME
            Returns dataframe with models and selected metrics score.

        '''

        models = getModels(mode='regression', modelClasses=['linear'])

        scoring = []
        for s in score:
            scoring.append(self._regression_scores[s]['name'])
            # results['test_' + regression_scores[s]['name']]= []

        results = self._runModels(models, x, y, scoring, cv)

        return pd.DataFrame(results)

    def treeRegression(self, x, y, score=['r2'], cv=5, **kwargs):
        '''
        Quick linear regression models evaluation.

        Parameters
        ----------
        x : DATAFRAME
            Dataset Features.
        y : NUMPY ARRAY
            Target.
        score : LIST, optional
            Scoring metrics. The default is ['r2'].
        cv : INT, optional
            Number of kfolds. The default is 5.

        Returns
        -------
        return : DATAFRAME
            Returns dataframe with models and selected metrics score.

        '''

        models = getModels(mode='regression', modelClasses=['tree'])

        scoring = []
        for s in score:
            scoring.append(self._regression_scores[s]['name'])
            # results['test_' + regression_scores[s]['name']]= []

        results = self._runModels(models, x, y, scoring, cv)

        return pd.DataFrame(results)

    def boostRegression(self, x, y, score=['r2'], cv=5, **kwargs):
        '''
        Quick linear regression models evaluation.

        Parameters
        ----------
        x : DATAFRAME
            Dataset Features.
        y : NUMPY ARRAY
            Target.
        score : LIST, optional
            Scoring metrics. The default is ['r2'].
        cv : INT, optional
            Number of kfolds. The default is 5.

        Returns
        -------
        return : DATAFRAME
            Returns dataframe with models and selected metrics score.

        '''

        models = getModels(mode='regression', modelClasses=['boost'])

        scoring = []
        for s in score:
            scoring.append(self._regression_scores[s]['name'])

        results = self._runModels(models, x, y, scoring, cv)

        return pd.DataFrame(results)

    def multiRegression(self, x, y, models=['boost', 'linear', 'tree'], score=['r2'], cv=5):

        models = getModels(mode='regression', modelClasses=models)

        scoring = []
        for s in score:
            scoring.append(self._regression_scores[s]['name'])

        results = self._runModels(models, x, y, scoring, cv)

        return pd.DataFrame(results)

    def customRegression(self, x, y, models=['boost', 'linear', 'tree'], selMode='type', score=['r2'], cv=5, model_selection='cross_validate', model_selection_params={}):

        models = getModels(mode='regression',
                           modelClasses=models, selMode=selMode)

        scoring = []
        for s in score:
            scoring.append(self._regression_scores[s]['name'])

        results = self._runModels(models, x, y, scoring, cv, model_selection=model_selection,
                                  metrics=score, model_selection_params=model_selection_params)

        return pd.DataFrame(results)

    def customRegressionYield(self, x, y, models=['boost', 'linear', 'tree'], selMode='type', score=['r2'], cv=5, model_selection='cross_validate', model_selection_params={}):

        models = getModels(mode='regression',
                           modelClasses=models, selMode=selMode)

        scoring = []
        for s in score:
            scoring.append(self._regression_scores[s]['name'])

        for i in self._runModelsYield(models, x, y, scoring, cv, model_selection=model_selection, metrics=score, model_selection_params=model_selection_params):
            yield i

# ============================================================
# Classification Functions
# ============================================================


class Classifier():

    def __init__(self):

        self._classification_scores = getScores('classification')

    def _crossvalidate(self, estimator, x, y, scoring, cv=5):

        result = cross_validate(
            estimator, x, y, scoring=scoring, cv=cv, n_jobs=-1)

        return result

    def _runModels(self, models, x, y, scoring, cv):
        results = {'estimator': [], 'fit_time': [], 'score_time': []}

        for model in models:
            res = self._crossvalidate(model['estimator'], x, y, scoring, cv)

            results['estimator'].append(model['name'])

            for k, v in res.items():
                if k not in results.keys():
                    results[k] = []

                results[k].append(v.mean())

        return results

    def svmClassifier(self, x, y, score=['accuracy'], cv=5, **kwargs):
        '''
        Quick SVM classification models evaluation.

        Parameters
        ----------
        x : DATAFRAME
            Dataset Features.
        y : NUMPY ARRAY
            Target.
        score : LIST, optional
            Scoring metrics. The default is ['accuracy'].
        cv : INT, optional
            Number of kfolds. The default is 5.

        Returns
        -------
        return : DATAFRAME
            Returns dataframe with models and selected metrics score.

        '''

        models = getModels(mode='classification', modelClasses=['svm'])

        scoring = []
        for s in score:
            scoring.append(self._classification_scores[s]['name'])

        results = self._runModels(models, x, y, scoring, cv)

        return pd.DataFrame(results)

    def treeClassifier(self, x, y, score=['accuracy'], cv=5, **kwargs):
        '''
        Quick Tree classification models evaluation.

        Parameters
        ----------
        x : DATAFRAME
            Dataset Features.
        y : NUMPY ARRAY
            Target.
        score : LIST, optional
            Scoring metrics. The default is ['accuracy'].
        cv : INT, optional
            Number of kfolds. The default is 5.

        Returns
        -------
        return : DATAFRAME
            Returns dataframe with models and selected metrics score.

        '''

        models = getModels(mode='classification', modelClasses=['tree'])

        scoring = []
        for s in score:
            scoring.append(self._classification_scores[s]['name'])

        results = self._runModels(models, x, y, scoring, cv)

        return pd.DataFrame(results)

    def boostClassifier(self, x, y, score=['accuracy'], cv=5, **kwargs):
        '''
        Quick Boost classification models evaluation.

        Parameters
        ----------
        x : DATAFRAME
            Dataset Features.
        y : NUMPY ARRAY
            Target.
        score : LIST, optional
            Scoring metrics. The default is ['accuracy'].
        cv : INT, optional
            Number of kfolds. The default is 5.

        Returns
        -------
        return : DATAFRAME
            Returns dataframe with models and selected metrics score.

        '''

        models = getModels(mode='classification', modelClasses=['boost'])

        scoring = []
        for s in score:
            scoring.append(self._classification_scores[s]['name'])

        results = self._runModels(models, x, y, scoring, cv)

        return pd.DataFrame(results)



# -*- coding: utf-8 -*-

'''
AutoTune models(:mod:`krakatoa.models.autotune`)
============================================================
'''

#============================================================
# Imports
#============================================================

from sklearn.model_selection import GridSearchCV
from ._getmodels import getModels
from ._metrics import getScores


import numpy as np
import math

from sklearn.utils._testing import ignore_warnings
from sklearn.exceptions import ConvergenceWarning

#============================================================
# Regression Functions
#============================================================

@ignore_warnings(category=ConvergenceWarning)
class RegressorAF():

    def __init__(self):
        self._regression_scores = getScores('regression')

    def fit(self, estimator, x, y, score=['r2'], cv=3, verbose=0, lr=0.001, random_state=0):
        results = {}
        modelConfig = getModels(mode='regression', modelClasses=[estimator], selMode='name', random_state=random_state)

        #Collector variables
        models = np.array([])
        opt_params = dict()
        scores = np.array([])
        
        if len(modelConfig) > 0:
            #Model definition
            # model = model[0]['estimator']
            modelConfig = modelConfig[0]
            model = modelConfig.get('estimator', '')

            scoring = []
            for s in score:
                scoring.append(self._regression_scores[s]['name'])
                            
            #carrega parametros
            # num_parameters = {key:attr['vals'] for key, attr in modelConfig['attr'].items() if attr['type'] == 'num' and attr['iterable'] == False}
            # num_parameters = {key:attr for key, attr in modelConfig['attr'].items() if attr['type'] == 'num' and attr['iterable'] == False}
            daux = {key:attr for key, attr in modelConfig['attr'].items() if attr['type'] == 'num'}

            # Alterado para que seja uma lista ordenada | pois dict nao pode ser |
            num_parameters = []
            for k, p in daux.items():
                p['key'] = k
                num_parameters.append(p)


            num_parameters = sorted(num_parameters, key= lambda item:item['iterable'])

            best_score = 0
            best_params = {}
            # for k, p in num_parameters.items(): 
            for item in num_parameters: 
 
                #Specific attr cfg
                # paramCfg = modelConfig['attr'][k]
                paramCfg = modelConfig['attr'][item['key']]

                stop = False
                #Process once to discover the best of initial three parameters
                grid_params =  {**{_k:np.array([_v]) for _k, _v in opt_params.items()}, item['key']:item['vals']}
                grid = GridSearchCV(model, param_grid=grid_params, cv=cv, n_jobs=-1, verbose=0)
                grid.fit(x, y)
                
                # Verifica se aplicando o grid search melhora resultados
                # se nao melhorar nao entra no pipeline para descobrir melhores parametros
                if grid.best_score_ > best_score:
                    
                    best_score = grid.best_score_
                    best_params = grid.best_params_
                    
                    last_best_model = grid.best_estimator_
                    last_best_score = best_score
                    last_best_param = best_params
                    
                else:
                    stop = True
                    
                if item['iterable'] == False:
                    param_range = item['vals']
                    
                    if verbose > 0:
                        print(f'-------- {item["key"]} ----------')
                    # TODO> Abstrair para um função             
                    while stop == False:
                        # if last_best_param[k] == None:
                        if last_best_param[item['key']] == None:
                            stop = True
                            break

                        # spec_best_param = last_best_param[k]
                        spec_best_param = last_best_param[item['key']]
                        
                        #identify if the value is in the center, left or right
                        index = np.where(param_range == spec_best_param)[0][0]
                        #left
                        if index == 0: 
                            lim_r = param_range[1]
                            lim_l = 0
                        elif index == 1:
                            lim_r = param_range[2]
                            lim_l = param_range[0]
                        else:
                            lim_r = spec_best_param*2 #TODO: Melhorar logica 
                            lim_l = param_range[1]

                        # Calculo dos novos limites a esquerda e direita    
                        aux_param_l = (spec_best_param - lim_l)/2
                        aux_param_r = (lim_r - spec_best_param)/2 + spec_best_param

                        param_l = aux_param_l if aux_param_l >= 0 else lim_l
                        param_r = aux_param_r if aux_param_r >= 0 else lim_r

                        # Verifica o tipo da variavel, se for INT devemos garantir que nao esteja passando como float e resolver

                        # if isinstance(paramCfg['ntype'], int):
                        if paramCfg['ntype'] == int:
                            param_l = int(math.ceil(param_l))
                            param_r = int(math.ceil(param_r))


                        #re do parameter range
                        param_range = np.array([param_l, spec_best_param , param_r])
                        
                        _grid_params =  {**{_k:np.array([_v]) for _k, _v in opt_params.items()}, item['key']:param_range}
                        _grid = GridSearchCV(model, param_grid=_grid_params, cv=cv, n_jobs=-1, verbose=0)
                        _grid.fit(x, y)
                        
                        _best_score = _grid.best_score_
                        _best_params = _grid.best_params_
                        _best_model = _grid.best_estimator_
                        
                        #  Show verbose if it is necessary
                        if verbose > 0:
                            
                            print("-----------------")
                            print("Range =====> ", param_range)
                            print("opt params: ", _grid_params)
                            print("Score: ", _best_score)
                            print("Params: ", _best_params)
                            print("Diff: ", (_best_score - last_best_score))
                            print("-----------------")
                            
                        if _best_score > last_best_score:
                            _diff = _best_score - last_best_score
                            
                            last_best_model = _best_model
                            last_best_score = _best_score
                            # last_best_param[k] = _best_params[k]
                            last_best_param[item['key']] = _best_params[item['key']]
                        
                            if _diff <= lr:
                                stop = True
                                
                        else:
                            stop = True

                # Trecho para iteraveis
                # TODO melhorar essa parte deixando mais robusto
                else:
                    if verbose > 0:
                        print(f'-------- {item["key"]} ----------')
                    min_try = 5 # Quantidade minima de tentativas  antes de sair do loop
                    value = item['init_val']
                    step = item['def_step']
                    maxVal = item['max']
                    hist_value = [] #tuple com value e accuracy

                    stop = False
                    try_count = 1

                    while stop == False:

                        _grid_params =  {**{_k:np.array([_v]) for _k, _v in opt_params.items()}, item['key']:[value]}
                        _grid = GridSearchCV(model, param_grid=_grid_params, cv=cv, n_jobs=-1, verbose=0)
                        _grid.fit(x, y)
                        
                        _best_score = _grid.best_score_
                        _best_params = _grid.best_params_
                        _best_model = _grid.best_estimator_
                        
                        print("_best_score")
                        print(_best_score)
                        hist_value.append((value, _best_score))
                        #  Show verbose if it is necessary
                        if verbose > 0:
                            
                            print("-----------------")
                            print("Value =====> ", value)
                            print("opt params: ", _grid_params)
                            print("Score: ", _best_score)
                            print("Params: ", _best_params)
                            print("Diff: ", (_best_score - last_best_score))
                            print("-----------------")
                            
                            
                        _diff = _best_score - last_best_score
                        
                        if last_best_score < _best_score:
                            last_best_model = _best_model
                            last_best_score = _best_score
                            # last_best_param[k] = _best_params[k]
                            last_best_param[item['key']] = _best_params[item['key']]
                    
                        if (_diff <= lr and try_count >= min_try) or (maxVal == value):
                            stop = True
                            print(hist_value)
                            # maxValue = max(hist_value, key=lambda item:item[1])
                        elif _diff >= lr and (min_try - try_count) < 2:
                            print("reset")
                            try_count = 0
                            min_try = 3 #Zera o contador e deixa pelo menor 3 a mais para tentar

                        try_count += 1
                        value += step

                
                #Store best params and results

                models = np.append(models, last_best_model)
                opt_params = {**opt_params, **last_best_param}
                scores = np.append(scores, last_best_score)
                    
        
        results = {
            'models' : models,
            'best_params' : opt_params,
            'scores' : scores
            }

        return results
    
    
class ClassifierAF():

    def __init__(self):
        self._classification_score = getScores('classification')

    def fit(self, estimator, x, y, score=['accuracy'], cv=3, verbose=0, lr=0.001, random_state=0):
        results = {}
        modelConfig = getModels(mode='classification', modelClasses=[estimator], selMode='name', random_state=random_state)

        #Collector variables
        models = np.array([])
        opt_params = dict()
        scores = np.array([])
        
        if len(modelConfig) > 0:
            #Model definition
            # model = model[0]['estimator']
            modelConfig = modelConfig[0]
            model = modelConfig.get('estimator', '')

            scoring = []
            for s in score:
                scoring.append(self._classification_score[s]['name'])
                            
            #carrega parametros
            # num_parameters = {key:attr['vals'] for key, attr in modelConfig['attr'].items() if attr['type'] == 'num' and attr['iterable'] == False}
            # num_parameters = {key:attr for key, attr in modelConfig['attr'].items() if attr['type'] == 'num' and attr['iterable'] == False}
            daux = {key:attr for key, attr in modelConfig['attr'].items() if attr['type'] == 'num'}

            # Alterado para que seja uma lista ordenada | pois dict nao pode ser |
            num_parameters = []
            for k, p in daux.items():
                p['key'] = k
                num_parameters.append(p)


            num_parameters = sorted(num_parameters, key= lambda item:item['iterable'])

            best_score = 0
            best_params = {}
            # for k, p in num_parameters.items(): 
            for item in num_parameters: 
 
                #Specific attr cfg
                # paramCfg = modelConfig['attr'][k]
                paramCfg = modelConfig['attr'][item['key']]

                stop = False
                #Process once to discover the best of initial three parameters
                grid_params =  {**{_k:np.array([_v]) for _k, _v in opt_params.items()}, item['key']:item['vals']}
                grid = GridSearchCV(model, param_grid=grid_params, cv=cv, n_jobs=-1, verbose=0)
                grid.fit(x, y)
                
                # Verifica se aplicando o grid search melhora resultados
                # se nao melhorar nao entra no pipeline para descobrir melhores parametros
                if grid.best_score_ > best_score:
                    
                    best_score = grid.best_score_
                    best_params = grid.best_params_
                    
                    last_best_model = grid.best_estimator_
                    last_best_score = best_score
                    last_best_param = best_params
                    
                else:
                    stop = True
                    
                if item['iterable'] == False:
                    param_range = item['vals']
                    
                    if verbose > 0:
                        print(f'-------- {item["key"]} ----------')
                        
                    # TODO> Abstrair para um função             
                    while stop == False:
                        # if last_best_param[k] == None:
                        if last_best_param[item['key']] == None:
                            stop = True
                            break

                        # spec_best_param = last_best_param[k]
                        spec_best_param = last_best_param[item['key']]
                        
                        #identify if the value is in the center, left or right
                        index = np.where(param_range == spec_best_param)[0][0]
                        #left
                        if index == 0: 
                            lim_r = param_range[1]
                            lim_l = 0
                        elif index == 1:
                            lim_r = param_range[2]
                            lim_l = param_range[0]
                        else:
                            lim_r = spec_best_param*2 #TODO: Melhorar logica 
                            lim_l = param_range[1]

                        # Calculo dos novos limites a esquerda e direita    
                        aux_param_l = (spec_best_param - lim_l)/2
                        aux_param_r = (lim_r - spec_best_param)/2 + spec_best_param

                        param_l = aux_param_l if aux_param_l >= 0 else lim_l
                        param_r = aux_param_r if aux_param_r >= 0 else lim_r

                        # Verifica o tipo da variavel, se for INT devemos garantir que nao esteja passando como float e resolver

                        # if isinstance(paramCfg['ntype'], int):
                        if paramCfg['ntype'] == int:
                            param_l = int(math.ceil(param_l))
                            param_r = int(math.ceil(param_r))


                        #re do parameter range
                        param_range = np.array([param_l, spec_best_param , param_r])
                        
                        _grid_params =  {**{_k:np.array([_v]) for _k, _v in opt_params.items()}, item['key']:param_range}
                        _grid = GridSearchCV(model, param_grid=_grid_params, cv=cv, n_jobs=-1, verbose=0)
                        _grid.fit(x, y)
                        
                        _best_score = _grid.best_score_
                        _best_params = _grid.best_params_
                        _best_model = _grid.best_estimator_
                        
                        #  Show verbose if it is necessary
                        if verbose > 0:
                            
                            print("-----------------")
                            print("Range =====> ", param_range)
                            print("opt params: ", _grid_params)
                            print("Score: ", _best_score)
                            print("Params: ", _best_params)
                            print("Diff: ", (_best_score - last_best_score))
                            print("-----------------")
                            
                        if _best_score > last_best_score:
                            _diff = _best_score - last_best_score
                            
                            last_best_model = _best_model
                            last_best_score = _best_score
                            # last_best_param[k] = _best_params[k]
                            last_best_param[item['key']] = _best_params[item['key']]
                        
                            if _diff <= lr:
                                stop = True
                                
                        else:
                            stop = True

                # Trecho para iteraveis
                # TODO melhorar essa parte deixando mais robusto
                else:
                    if verbose > 0:
                        print(f'-------- {item["key"]} ----------')

                    min_try = 5 # Quantidade minima de tentativas  antes de sair do loop
                    value = item['init_val']
                    step = item['def_step']
                    maxVal = item['max']
                    hist_value = [] #tuple com value e accuracy

                    stop = False
                    try_count = 1

                    while stop == False:

                        _grid_params =  {**{_k:np.array([_v]) for _k, _v in opt_params.items()}, item['key']:[value]}
                        _grid = GridSearchCV(model, param_grid=_grid_params, cv=cv, n_jobs=-1, verbose=0)
                        _grid.fit(x, y)
                        
                        _best_score = _grid.best_score_
                        _best_params = _grid.best_params_
                        _best_model = _grid.best_estimator_
                        
                        print("_best_score")
                        print(_best_score)
                        hist_value.append((value, _best_score))
                        #  Show verbose if it is necessary
                        if verbose > 0:
                            
                            print("-----------------")
                            print("Value =====> ", value)
                            print("opt params: ", _grid_params)
                            print("Score: ", _best_score)
                            print("Params: ", _best_params)
                            print("Diff: ", (_best_score - last_best_score))
                            print("-----------------")
                            
                            
                        _diff = _best_score - last_best_score
                        
                        if last_best_score < _best_score:
                            last_best_model = _best_model
                            last_best_score = _best_score
                            # last_best_param[k] = _best_params[k]
                            last_best_param[item['key']] = _best_params[item['key']]
                    
                        if (_diff <= lr and try_count >= min_try) or (maxVal == value):
                            stop = True
                            print(hist_value)
                            # maxValue = max(hist_value, key=lambda item:item[1])
                        elif _diff >= lr and (min_try - try_count) < 2:
                            print("reset")
                            try_count = 0
                            min_try = 3 #Zera o contador e deixa pelo menor 3 a mais para tentar

                        try_count += 1
                        value += step

                
                #Store best params and results

                models = np.append(models, last_best_model)
                opt_params = {**opt_params, **last_best_param}
                scores = np.append(scores, last_best_score)
                    
        
        results = {
            'models' : models,
            'best_params' : opt_params,
            'scores' : scores
            }

        return results
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    



# -*- coding: utf-8 -*-

'''
Configuration file(:mod:`krakatoa._config`)
============================================================
'''

# ============================================================
# Imports
# ============================================================
from pickle import NONE
from telnetlib import KERMIT
from sklearn.linear_model import LinearRegression, Ridge, Lasso, ElasticNet
from sklearn.ensemble import RandomForestRegressor, AdaBoostRegressor
from sklearn.tree import DecisionTreeRegressor, ExtraTreeRegressor
from xgboost import XGBRegressor

from sklearn.svm import SVC
from sklearn.ensemble import RandomForestClassifier, ExtraTreesClassifier, AdaBoostClassifier
from sklearn.tree import DecisionTreeClassifier
from xgboost import XGBClassifier

import numpy as np

# ============================================================
# Configuration
# ============================================================
def getConfigRegr():
    
    conf = {
        'LINEAR': {
            'type': 'linear',
            'name': 'linear',
            'estimator': LinearRegression(),
            'attr': {}
        },
        'RIDGE': {
            'type': 'linear',
            'name': 'ridge',
            'estimator': Ridge(),
            'attr': {
                'alpha': {
                    'vals': np.array([0.1, 1, 10]),
                    'iterable': False,
                    'type': 'num',
                    'ntype' : float,
                    'min': 0,
                    'max': None
                }
            },
        },
        'LASSO': {
            'type': 'linear',
            'name': 'lasso',
            'estimator': Lasso(),
            'attr': {
                'alpha': {
                    'vals': np.array([0.1, 1, 10]),
                    'iterable': False,
                    'type': 'num',
                    'ntype' : float,
                    'min': 0,
                    'max': None
                },
                'tol': {
                    'vals': np.array([0.001, 0.01, 0.1]),
                    'iterable': False,
                    'type': 'num',
                    'ntype' : float,
                    'min': 0,
                    'max': None
                }

            },
        },
        'ELASTIC_NET': {
            'type': 'linear',
            'name': 'elasticnet',
            'estimator': ElasticNet(),
            'attr': {
                'alpha': {
                    'vals': np.array([0.1, 1, 10]),
                    'iterable': False,
                    'type': 'num',
                    'ntype' : float,
                    'min': 0,
                    'max': None
                },
                'l1_ratio': {
                    'vals': np.array([0.1, 0.5, 0.9]),
                    'iterable': False,
                    'ntype' : float,
                    'type': 'num',
                    'min': 0,
                    'max': 1
                },
                'tol': {
                    'vals': np.array([0.001, 0.01, 0.1]),
                    'iterable': False,
                    'ntype' : float,
                    'type': 'num',
                    'min': 0,
                    'max': None
                }
            },
        },
        'DECISION_TREE': {
            'type': 'tree',
            'name': 'decisionTree',
            'estimator': DecisionTreeRegressor(),
            'attr': {
                'min_samples_split': {
                    'vals': np.array([0.1, 0.5, 1]),
                    'iterable': False,
                    'type': 'num',
                    'ntype' : float,
                    'min': 0.1,
                    'max': 1
                },
                'min_samples_leaf': {
                    'vals': np.array([0.1, 1, 5]),
                    'iterable': False,
                    'type': 'num',
                    'ntype' : float,
                    'min': 0.1,
                    'max': None
                },
                'min_weight_fraction_leaf': {
                    'vals': np.array([0.0, 0.25, 0.5]),
                    'iterable': False,
                    'type': 'num',
                    'ntype' : float,
                    'min': 0,
                    'max': None
                },
                'min_impurity_decrease': {
                    'vals': np.array([0.0, 0.5, 1.0]),
                    'iterable': False,
                    'type': 'num',
                    'ntype' : float,
                    'min': 0,
                    'max': None
                },
                'max_depth': {
                    'vals': np.array([1, 5, 10]),
                    'iterable': False,
                    'type': 'num',
                    'ntype' : int,
                    'min': 1,
                    'max': None
                }
            },
        },
        'RANDOM_FOREST': {
            'type': 'tree',
            'name': 'randomForest',
            'estimator': RandomForestRegressor(),
            'attr': {
                'n_estimators': {
                    'vals': [100],
                    'iterable': True,
                    'init_val': 100,
                    'def_step': 100,
                    'type': 'num',
                    'ntype' : int,
                    'min': 5,
                    'max': None
                },
                'min_samples_split': {
                    'vals': np.array([0.1, 0.5, 1]),
                    'iterable': False,
                    'type': 'num',
                    'ntype' : float,
                    'min': 0.1,
                    'max': None
                },
                'min_samples_leaf': {
                    'vals': np.array([0.1, 1, 5]),
                    'iterable': False,
                    'type': 'num',
                    'ntype' : float,
                    'min': 0.1,
                    'max': None
                },
                'min_weight_fraction_leaf': {
                    'vals': np.array([0.0, 0.25, 0.5]),
                    'iterable': False,
                    'type': 'num',
                    'ntype' : float,
                    'min': 0,
                    'max': None
                },
                'min_impurity_decrease': {
                    'vals': np.array([0.0, 0.5, 1.0]),
                    'iterable': False,
                    'type': 'num',
                    'ntype' : float,
                    'min': 0,
                    'max': None
                },
                'max_depth': {
                    'vals': np.array([1, 3, 5]),
                    'iterable': False,
                    'type': 'num',
                    'ntype' : int,
                    'min': 0,
                    'max': None
                }
            },
        },
        'EXTRA_TREE': {
            'type': 'tree',
            'name': 'extraTree',
            'estimator': ExtraTreeRegressor(),
            'attr': {
                # 'n_estimators': {
                #     'vals': [100],
                #     'iterable': True,
                #     'init_val': 100,
                #     'def_step': 100,
                #     'type': 'num',
                #     'ntype' : int,
                #     'min': 5,
                #     'max': None
                # },
                'min_samples_split': {
                    'vals': np.array([0.1, 0.5, 1]),
                    'iterable': False,
                    'type': 'num',
                    'ntype' : float,
                    'min': 0.1,
                    'max': None
                },
                'min_samples_leaf': {
                    'vals': np.array([0.1, 1, 5]),
                    'iterable': False,
                    'type': 'num',
                    'ntype' : float,
                    'min': 0.1,
                    'max': None
                },
                'min_weight_fraction_leaf': {
                    'vals': np.array([0.0, 0.25, 0.5]),
                    'iterable': False,
                    'type': 'num',
                    'ntype' : float,
                    'min': 0,
                    'max': None
                },
                'min_impurity_decrease': {
                    'vals': np.array([0.0, 0.5, 1.0]),
                    'iterable': False,
                    'type': 'num',
                    'ntype' : float,
                    'min': 0,
                    'max': None
                },
                'max_depth': {
                    'vals': np.array([1, 5, 10]),
                    'iterable': False,
                    'type': 'num',
                    'ntype' : int,
                    'min': 1,
                    'max': None
                }
            },
        },
        'ADABOOST': {
            'type': 'boost',
            'name': 'adaboost',
            'estimator': AdaBoostRegressor(),
            'attr': {
                'n_estimators': {
                    'vals': [50],
                    'iterable': True,
                    'init_val': 50,
                    'def_step': 50,
                    'type': 'num',
                    'ntype' : int,
                    'min': 5,
                    'max': None
                },
                'learning_rate': {
                    'vals': np.array([0.001, 0.1, 1]),
                    'iterable': False,
                    'type': 'num',
                    'ntype' : float,
                    'min': 0.0,
                    'max': None
                },
                'loss': {
                    'vals': ['linear', 'square', 'exponential'],
                    'iterable': False,
                    'type': 'cat',
                    'min': 0.0,
                    'max': None
                }
            }
        },
        'XGBOOST': {
            'type': 'boost',
            'name': 'xgboost',
            'estimator': XGBRegressor(),
            'attr': {
                'eta': {
                    'vals': np.array([0.1, 0.3, 0.5]),
                    'default': 0.3,
                    'iterable': False,
                    'type': 'num',
                    'ntype' : float,
                    'min': 0.0,
                    'max': 1.0
                },
                'gamma': {
                    'vals': np.array([0, 1, 10]),
                    'default': 0,
                    'iterable': False,
                    'type': 'num',
                    'ntype' : int,
                    'min': 0,
                    'max': None
                },
                'max_depth': {
                    'vals': np.array([2, 6, 15]),
                    'default': 6,
                    'iterable': False,
                    'type': 'num',
                    'ntype' : int,
                    'min': 0,
                    'max': None
                },
                'min_child_weight': {
                    'vals': np.array([0, 5, 10]),
                    'default': 1,
                    'iterable': False,
                    'type': 'num',
                    'ntype' : int,
                    'min': 0.0,
                    'max': None
                },
                'max_delta_step': {
                    'vals': np.array([0, 5, 10]),
                    'default': 0,
                    'iterable': False,
                    'type': 'num',
                    'ntype' : int,                   
                    'min': 0.0,
                    'max': None
                },
                'subsample': {
                    'vals': np.array([0, 0.5, 1]),
                    'default': 1,
                    'iterable': False,
                    'type': 'num',
                    'ntype' : int,
                    'min': 0.0,
                    'max': 1.0
                },
                'lambda': {  # L2 regularizarion
                    'vals': np.array([1, 10, 100]),
                    'default': 1,
                    'iterable': False,
                    'type': 'num',
                    'ntype' : int,
                    'min': 0,
                    'max': None
                },
                'alpha': {  # L1 regularizarion
                    'vals': np.array([0, 10, 100]),
                    'default': 0,
                    'iterable': False,
                    'type': 'num',
                    'ntype' : int,
                    'min': 0,
                    'max': None
                },
                'max_leaves': {
                    'vals': np.array([0, 5, 10]),
                    'default': 0,
                    'iterable': False,
                    'type': 'num',
                    'ntype' : int,
                    'min': 0,
                    'max': None
                },
                'loss': {
                    'vals': ['linear', 'square', 'exponential'],
                    'iterable': False,
                    'type': 'cat',
                    'min': 0.0,
                    'max': None
                }
            },
        }
    }

    return conf

def getConfigClass():
    
    conf = {
        'SVC_LINEAR': {
            'type': 'svm',
            'name': 'svc_linear',
            'estimator': SVC(kernel='linear'),
            'attr': {
                'C': {
                    'vals': np.array([0.1, 1, 10]),
                    'iterable': False,
                    'type': 'num',
                    'ntype' : float,
                    'min': 0.1,
                    'max': None
                },
                'tol': {
                    'vals': np.array([0.00001, 0.0001, 0.01]),
                    'iterable': False,
                    'type': 'num',
                    'ntype' : float,
                    'min': 0.000000001,
                    'max': None
                },
                'max_iter': {
                    'vals': [500],
                    'iterable': True,
                    'init_val': 500,
                    'def_step': 100,
                    'type': 'num',
                    'ntype' : int,
                    'min': 100,
                    'max': None
                }
            }
        },
        'SVC_RBF': {
            'type': 'svm',
            'name': 'svc_rbf',
            'estimator': SVC(kernel='rbf'),
            'attr': {
                'C': {
                    'vals': np.array([0.1, 1, 10]),
                    'iterable': False,
                    'type': 'num',
                    'ntype' : float,
                    'min': 0.1,
                    'max': None
                },
                'tol': {
                    'vals': np.array([0.00001, 0.0001, 0.01]),
                    'iterable': False,
                    'type': 'num',
                    'ntype' : float,
                    'min': 0.000000001,
                    'max': None
                },
                'max_iter': {
                    'vals': [500],
                    'iterable': True,
                    'init_val': 500,
                    'def_step': 100,
                    'type': 'num',
                    'ntype' : int,
                    'min': 100,
                    'max': None
                }
            }
        },
        'SVC_POLY': {
            'type': 'svm',
            'name': 'svc_poly',
            'estimator': SVC(kernel='poly'),
            'attr': {
                'C': {
                    'vals': np.array([0.1, 1, 10]),
                    'iterable': False,
                    'type': 'num',
                    'ntype' : float,
                    'min': 0.1,
                    'max': None
                },
                'degree' : {
                    'vals': [2],
                    'iterable': True,
                    'init_val': 2,
                    'def_step': 1,
                    'type': 'num',
                    'ntype' : int,
                    'min': 2,
                    'max': 5
                },
                'coef0': {
                    'vals': np.array([0.0, 0.5, 1.0]),
                    'iterable': False,
                    'type': 'num',
                    'ntype' : float,
                    'min': 0.0,
                    'max': None
                },
                'tol': {
                    'vals': np.array([0.00001, 0.0001, 0.01]),
                    'iterable': False,
                    'type': 'num',
                    'ntype' : float,
                    'min': 0.000000001,
                    'max': None
                },
                'max_iter': {
                    'vals': [500],
                    'iterable': True,
                    'init_val': 500,
                    'def_step': 100,
                    'type': 'num',
                    'ntype' : int,
                    'min': 100,
                    'max': None
                }
            }
        },
        'SVC_SIGMOID': {
            'type': 'svm',
            'name': 'svc_sigmoid',
            'estimator': SVC(kernel='sigmoid'),
            'attr': {
                'C': {
                    'vals': np.array([0.1, 1, 10]),
                    'iterable': False,
                    'type': 'num',
                    'ntype' : float,
                    'min': 0.1,
                    'max': None
                },
                'degree' : {
                    'vals': [2],
                    'iterable': True,
                    'init_val': 2,
                    'def_step': 1,
                    'type': 'num',
                    'ntype' : int,
                    'min': 2,
                    'max': 5
                },
                'coef0': {
                    'vals': np.array([0.0, 0.5, 1.0]),
                    'iterable': False,
                    'type': 'num',
                    'ntype' : float,
                    'min': 0.0,
                    'max': None
                },
                'tol': {
                    'vals': np.array([0.00001, 0.0001, 0.01]),
                    'iterable': False,
                    'type': 'num',
                    'ntype' : float,
                    'min': 0.000000001,
                    'max': None
                },
                'max_iter': {
                    'vals': [500],
                    'iterable': True,
                    'init_val': 500,
                    'def_step': 100,
                    'type': 'num',
                    'ntype' : int,
                    'min': 100,
                    'max': None
                }
            }
        },
        'RANDOM_FOREST': {
            'type': 'tree',
            'name': 'randomForest',
            'estimator': RandomForestClassifier(),
            'attr': {
                'n_estimators': {
                    'vals': [100],
                    'iterable': True,
                    'init_val': 100,
                    'def_step': 100,
                    'type': 'num',
                    'ntype' : int,
                    'min': 5,
                    'max': None
                },
                'min_samples_split': {
                    'vals': np.array([0.1, 0.5, 1]),
                    'iterable': False,
                    'type': 'num',
                    'ntype' : float,
                    'min': 0.1,
                    'max': None
                },
                'min_samples_leaf': {
                    'vals': np.array([0.1, 1, 5]),
                    'iterable': False,
                    'type': 'num',
                    'ntype' : float,
                    'min': 0.1,
                    'max': None
                },
                'min_weight_fraction_leaf': {
                    'vals': np.array([0.0, 0.25, 0.5]),
                    'iterable': False,
                    'type': 'num',
                    'ntype' : float,
                    'min': 0,
                    'max': None
                },
                'min_impurity_decrease': {
                    'vals': np.array([0.0, 0.5, 1.0]),
                    'iterable': False,
                    'type': 'num',
                    'ntype' : float,
                    'min': 0,
                    'max': None
                },
                'max_depth': {
                    'vals': np.array([1, 3, 5]),
                    'iterable': False,
                    'type': 'num',
                    'ntype' : int,
                    'min': 0,
                    'max': None
                }
            },
        },
        'EXTRA_TREE': {
            'type': 'tree',
            'name': 'extraTree',
            'estimator': ExtraTreesClassifier(),
            'attr': {
                'n_estimators': {
                    'vals': [100],
                    'iterable': True,
                    'init_val': 100,
                    'def_step': 100,
                    'type': 'num',
                    'ntype' : int,
                    'min': 5,
                    'max': None
                },
                'min_samples_split': {
                    'vals': np.array([0.1, 0.5, 1]),
                    'iterable': False,
                    'type': 'num',
                    'ntype' : float,
                    'min': 0.1,
                    'max': None
                },
                'min_samples_leaf': {
                    'vals': np.array([0.1, 1, 5]),
                    'iterable': False,
                    'type': 'num',
                    'ntype' : float,
                    'min': 0.1,
                    'max': None
                },
                'min_weight_fraction_leaf': {
                    'vals': np.array([0.0, 0.25, 0.5]),
                    'iterable': False,
                    'type': 'num',
                    'ntype' : float,
                    'min': 0,
                    'max': None
                },
                'min_impurity_decrease': {
                    'vals': np.array([0.0, 0.5, 1.0]),
                    'iterable': False,
                    'type': 'num',
                    'ntype' : float,
                    'min': 0,
                    'max': None
                },
                'max_depth': {
                    'vals': np.array([1, 5, 10]),
                    'iterable': False,
                    'type': 'num',
                    'ntype' : int,
                    'min': 1,
                    'max': None
                }
            },
        },
        'DECISION_TREE': {
            'type': 'tree',
            'name': 'decisionTree',
            'estimator': DecisionTreeClassifier(),
            'attr': {
                'min_samples_split': {
                    'vals': np.array([0.1, 0.5, 1]),
                    'iterable': False,
                    'type': 'num',
                    'ntype' : float,
                    'min': 0.1,
                    'max': 1
                },
                'min_samples_leaf': {
                    'vals': np.array([0.1, 1, 5]),
                    'iterable': False,
                    'type': 'num',
                    'ntype' : float,
                    'min': 0.1,
                    'max': None
                },
                'min_weight_fraction_leaf': {
                    'vals': np.array([0.0, 0.25, 0.5]),
                    'iterable': False,
                    'type': 'num',
                    'ntype' : float,
                    'min': 0,
                    'max': None
                },
                'min_impurity_decrease': {
                    'vals': np.array([0.0, 0.5, 1.0]),
                    'iterable': False,
                    'type': 'num',
                    'ntype' : float,
                    'min': 0,
                    'max': None
                },
                'max_depth': {
                    'vals': np.array([1, 5, 10]),
                    'iterable': False,
                    'type': 'num',
                    'ntype' : int,
                    'min': 1,
                    'max': None
                }
            },
        },
        'ADABOOST': {
            'type': 'boost',
            'name': 'adaboost',
            'estimator': AdaBoostClassifier(),
            'attr': {
                'n_estimators': {
                    'vals': [50],
                    'iterable': True,
                    'init_val': 50,
                    'def_step': 50,
                    'type': 'num',
                    'ntype' : int,
                    'min': 5,
                    'max': None
                },
                'learning_rate': {
                    'vals': np.array([0.001, 0.1, 1]),
                    'iterable': False,
                    'type': 'num',
                    'ntype' : float,
                    'min': 0.0,
                    'max': None
                },
                'loss': {
                    'vals': ['linear', 'square', 'exponential'],
                    'iterable': False,
                    'type': 'cat',
                    'min': 0.0,
                    'max': None
                }
            }
        },
        'XGBOOST': {
            'type': 'boost',
            'name': 'xgboost',
            'estimator': XGBClassifier(),
            'attr': {
                'eta': {
                    'vals': np.array([0.1, 0.3, 0.5]),
                    'default': 0.3,
                    'iterable': False,
                    'type': 'num',
                    'ntype' : float,
                    'min': 0.0,
                    'max': 1.0
                },
                'gamma': {
                    'vals': np.array([0, 1, 10]),
                    'default': 0,
                    'iterable': False,
                    'type': 'num',
                    'ntype' : int,
                    'min': 0,
                    'max': None
                },
                'max_depth': {
                    'vals': np.array([2, 6, 15]),
                    'default': 6,
                    'iterable': False,
                    'type': 'num',
                    'ntype' : int,
                    'min': 0,
                    'max': None
                },
                'min_child_weight': {
                    'vals': np.array([0, 5, 10]),
                    'default': 1,
                    'iterable': False,
                    'type': 'num',
                    'ntype' : int,
                    'min': 0.0,
                    'max': None
                },
                'max_delta_step': {
                    'vals': np.array([0, 5, 10]),
                    'default': 0,
                    'iterable': False,
                    'type': 'num',
                    'ntype' : int,                   
                    'min': 0.0,
                    'max': None
                },
                'subsample': {
                    'vals': np.array([0, 0.5, 1]),
                    'default': 1,
                    'iterable': False,
                    'type': 'num',
                    'ntype' : int,
                    'min': 0.0,
                    'max': 1.0
                },
                'lambda': {  # L2 regularizarion
                    'vals': np.array([1, 10, 100]),
                    'default': 1,
                    'iterable': False,
                    'type': 'num',
                    'ntype' : int,
                    'min': 0,
                    'max': None
                },
                'alpha': {  # L1 regularizarion
                    'vals': np.array([0, 10, 100]),
                    'default': 0,
                    'iterable': False,
                    'type': 'num',
                    'ntype' : int,
                    'min': 0,
                    'max': None
                },
                'max_leaves': {
                    'vals': np.array([0, 5, 10]),
                    'default': 0,
                    'iterable': False,
                    'type': 'num',
                    'ntype' : int,
                    'min': 0,
                    'max': None
                },
                'loss': {
                    'vals': ['linear', 'square', 'exponential'],
                    'iterable': False,
                    'type': 'cat',
                    'min': 0.0,
                    'max': None
                }
            },
        }



    }

    return conf




# -*- coding: utf-8 -*-

'''
Experiment creation (:mod:`krakatoa.future.experiment`)
============================================================
'''

# ============================================================
# Imports
# ============================================================

import pandas as pd

from ..models.quick import Regressor  # Quick models regressor
from ..models.autotune import RegressorAF  # Auto models regressor
from .preprocess import DataClean, splitDataset

# ============================================================


class Experiment(DataClean):

    def __init__(self):
        super().__init__()
        self.dataset = pd.DataFrame()
        self.target = None

    def _setTarget(self, target):
        self.target = target

    def create(self, dataset, target=None):

        # set target if informed
        if target != None:
            self._setTarget(target)

        # datasetClass = DataClean()

        # Fast and simple method
        # datasetClass.fit_transform(dataset)
        super().fit_transform(dataset)

    def train(self, method='quick', estimator=['linear', 'tree'], **kwargs):

        X, y = super().splitTrainTest(self.target)

        if method == 'quick':
            regressor = Regressor()
            result = regressor.multiRegression(
                x=X, y=y, models=estimator, **kwargs)
        elif method == 'auto':
            regressor = RegressorAF()
            result = regressor.fit(estimator=estimator, x=X, y=y, **kwargs)

        return result



# -*- coding: utf-8 -*-

'''
Prediction module (:mod:`krakatoa.future.predict`)
============================================================
'''

# ============================================================
# Imports
# ============================================================
import pickle
import joblib
import numpy as np
# ============================================================


class Predictor():

    def __init__(self):
        self.model = None

    def loadModel(self, model=None, model_path = None, loader_engine: str = 'pickle'):

        if model is None:

            if loader_engine == 'pickle':
                with open(model_path, 'rb') as handle:
                    model = pickle.load(handle)

            elif loader_engine == 'joblib':

                model = joblib.load(model_path)
                
            else:
                raise ValueError('The selected loader_engine is not valid! Please select one of the following: pickle, joblib')

        self.model = model
       
    def predict(self, X: np.array):
        
        # Reshape X data
        _X = X.reshape(-1, X.shape[0])
        _prediction = self.model.predict(_X)[0]
        
        return _prediction



# -*- coding: utf-8 -*-

'''
Metrics file(:mod:`krakatoa.models._model_selection`)
============================================================
'''

#============================================================
# Imports
#============================================================

from sklearn import model_selection

#============================================================
# Metrics
#============================================================
#TODO: Criar modulo para metrics

def getModel(model):

    models = {
        'train_test_split' : {'f' : model_selection.train_test_split, 'name' : 'train_test_split', 'method' : 0},
        'KFold' : {'f' : model_selection.KFold, 'name' : 'KFold', 'method' : 1},
        # 'GroupKFold' : {'f' : model_selection.GroupKFold, 'name' : 'GroupKFold', 'method' : 1},
        'RepeatedKFold' : {'f' : model_selection.RepeatedKFold, 'name' : 'RepeatedKFold', 'method' : 1},
        'StratifiedGroupKFold' : {'f' : model_selection.StratifiedGroupKFold, 'name' : 'StratifiedGroupKFold', 'method' : 1},
        'cross_validate' : {'f' : model_selection.cross_validate, 'name' : 'cross_validate', 'method' : 2},
        'cross_val_score' : {'f' : model_selection.cross_val_score, 'name' : 'cross_val_score', 'method' : 2},
        'build' : {'f' : None, 'name' : None, 'method' : 3}
        }
    
    return models.get(model, None)